/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Ant-Contrib project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Ant-Contrib project (http://sourceforge.net/projects/ant-contrib)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name Ant-Contrib must not be used to endorse or promote products
 *    derived from this software without prior written permission. For
 *    written permission, please contact
 *    ant-contrib-developers@lists.sourceforge.net.
 *
 * 5. Products derived from this software may not be called "Ant-Contrib"
 *    nor may "Ant-Contrib" appear in their names without prior written
 *    permission of the Ant-Contrib project.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE ANT-CONTRIB PROJECT OR ITS
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */

package net.sf.antcontrib.logic;

import java.io.*;
import java.util.*;

import org.apache.tools.ant.*;
import org.apache.tools.ant.taskdefs.*;
import org.apache.tools.ant.taskdefs.condition.*;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.util.*;
import org.apache.tools.ant.types.PatternSet;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.types.Mapper;

/**
* Task to help in calling tasks if generated files are older
* than source files.
* Sets a given property or runs an internal task.
*
* Based on
*  org.apache.org.apache.tools.ant.taskdefs.UpToDate
*
* @author peter reilly
*/

public class OutOfDate extends Task implements Condition {


    private TaskContainer doTask = null;
    private String property;
    private String value = "true";
    private Vector mappers = new Vector();
    private Path targetpaths = null;
    private Path sourcepaths = null;
    private String outputSources = null;
    private String outputSourcesPath = null;
    private String outputTargets = null;
    private String outputTargetsPath = null;
    private String allTargets = null;
    private String separator = " ";
    
    private Hashtable targetSet = new Hashtable();
    private Hashtable sourceSet = new Hashtable();
    private Hashtable allTargetSet = new Hashtable();

    /**
     * Defines the FileNameMapper to use (nested mapper element).
     */
    public Mapper createMapper() {
        Mapper mapper = new Mapper(getProject());
        mappers.addElement(mapper);
        return mapper;
    }

    /**
     * The property to set if any of the target files are outofdate with
     * regard to any of the source files.
     *
     * @param property the name of the property to set if Target is outofdate.
     */
    public void setProperty(String property) {
        this.property = property;
    }

    /**
     * The separator to use to separate the files
     */

    public void setSeparator(String separator) {
        this.separator = separator;
    }

    /**
     * The value to set the named property to the target files
     * are outofdate
     *
     * @param value the value to set the property
     */
    public void setValue(String value) {
        this.value = value;
    }

    public Path createTargetfiles()
    {
        if (targetpaths == null) {
            targetpaths = new Path(getProject());
        }
        return targetpaths;
    }

    public Path createSourcefiles()
    {
        if (sourcepaths == null) {
            sourcepaths = new Path(getProject());
        }
        return sourcepaths;
    }

    public void setOutputSources(String outputSources) {
        this.outputSources = outputSources;
    }

    public void setOutputTargets(String outputTargets) {
        this.outputTargets = outputTargets;
    }

    public void setOutputTargetsPath(String outputTargetsPath) {
        this.outputTargetsPath = outputTargetsPath;
    }

    public void setAllTargets(String allTargets) {
        this.allTargets = allTargets;
    }

    public void setOutputSourcesPath(String outputSourcesPath) {
        this.outputSourcesPath = outputSourcesPath;
    }


    /**
     * Embedded do sequential.
     */
    public void addParallel(Parallel doTask) {
        if (this.doTask != null) {
            throw new BuildException(
                "You must not nest more that one <parallel> or <sequential>" +
                " into <outofdate>");
        }
        this.doTask = doTask;
    }

    /**
     * Embedded do sequential.
     */
    public void addSequential(Sequential doTask) {
        if (this.doTask != null) {
            throw new BuildException(
                "You must not nest more that one <parallel> or <sequential>" +
                " into <outofdate>");
        }
        this.doTask = doTask;
    }

    /**
     * Evaluate (all) target and source file(s) to
     * see if the target(s) is/are outoutdate.
     */
    public boolean eval() {
        boolean ret = false;
        FileUtils fileUtils = FileUtils.newFileUtils();
        if (sourcepaths == null)
            throw new BuildException(
                "You must specify a <sourcefiles> element.");

        if (targetpaths == null && mappers.size() == 0)
            throw new BuildException(
                "You must specify a <targetfiles> or <mapper> element.");

        // Source Paths
        String spaths[] = sourcepaths.list();
        String baseDir;
        File   baseDirFile;
        int    baseDirLength;
        baseDir = getProject().getProperty("basedir");
        baseDirFile = new File(baseDir);
        baseDirLength = baseDir.length();
        if (baseDir.charAt(baseDir.length() - 1) != File.separatorChar)
            baseDir = baseDir + File.separator;

        String rpaths[] = new String[spaths.length];
        for (int i = 0; i < spaths.length; ++i) {
            int pos = spaths[i].indexOf(baseDir);
            if (pos != 0)
                rpaths[i] = spaths[i];
            else
                rpaths[i] = spaths[i].substring(baseDirLength+1);
        }
        

        for (int i = 0; i < rpaths.length; ++i)
            System.out.println(rpaths[i]);

        for (int i=0;i<spaths.length;i++) {
            File sourceFile = new File(spaths[i]);
            if (! sourceFile.exists()) {
                throw new BuildException(sourceFile.getAbsolutePath()
                                         + " not found.");
            }
        }

        // Target Paths

        if (targetpaths != null) {
            String[] paths = targetpaths.list();
            for (int i = 0; i < paths.length; ++i) {
                if (targetNeedsGen(paths[i], spaths)) {
                    ret = true;
                }
            }
        }

        // Mapper Paths
        for (Enumeration e = mappers.elements(); e.hasMoreElements();)
        {
            FileNameMapper fileNameMapper =
                ((Mapper) e.nextElement()).getImplementation();
            Vector mappedFiles = new Vector();
            for (int i = 0; i < spaths.length; ++i) {
                String mapped[] = fileNameMapper.mapFileName(rpaths[i]);
                if (mapped != null) {
                    for (int j = 0; j < mapped.length; ++j) {
                        if (outOfDate(
                                new File(spaths[i]),
                                fileUtils.resolveFile(
                                    baseDirFile, mapped[j])))
                        {
                            ret = true;
                        }
                    }
                }
            }
        }

        if (allTargets != null)
            this.getProject().setNewProperty(
                allTargets, setToString(allTargetSet));

        if (outputSources != null)
            this.getProject().setNewProperty(
                outputSources, setToString(sourceSet));

        if (outputTargets != null)
            this.getProject().setNewProperty(
                outputTargets, setToString(targetSet));

        if (outputSourcesPath != null)
            this.getProject().addReference(
                outputSourcesPath, setToPath(sourceSet));

        if (outputTargetsPath != null)
            this.getProject().addReference(
                outputTargetsPath, setToPath(targetSet));

        if (ret) {
            if (property != null)
                this.getProject().setNewProperty(property, value);
        }

        return ret;
    }

    private boolean targetNeedsGen(String target, String spaths[]) {
        boolean ret = false;
        File targetFile = new File(target);
        for (int i=0;i<spaths.length;i++) {
            if (outOfDate(new File(spaths[i]), targetFile))
                ret = true;
        }
        return ret;
    }

    /**
     * Sets property to true and/or executes embedded do
     * if any of the target file(s) do not have a more recent timestamp
     * than (each of) the source file(s).
     */
    public void execute() throws BuildException {
        if (property == null && doTask == null)
            throw new BuildException(
            "property attribute or do element is required.");

        if (! eval())
            return;

        if (property != null)
            this.getProject().setNewProperty(property, value);

        if (doTask != null)
        {
            if (doTask instanceof Parallel)
                ((Parallel)doTask).perform();
            else if (doTask instanceof Sequential)
                ((Sequential)doTask).perform();
            else
                throw new BuildException("Illegal do Task");
        }
    }


    private boolean outOfDate(File sourceFile, File targetFile) {
        boolean ret  = false;
        allTargetSet.put(targetFile, targetFile);
        if (! targetFile.exists())
            ret = true;
        if (! ret)
            ret = sourceFile.lastModified() > targetFile.lastModified();
        if (ret) {
            sourceSet.put(sourceFile, sourceFile);
            targetSet.put(targetFile, targetFile);
        }
        return ret;
    }

    private String setToString(Hashtable set) {
        StringBuffer b = new StringBuffer();
        for (Enumeration e = set.keys(); e.hasMoreElements();) {
            File v = (File) e.nextElement();
            if (b.length() != 0)
                b.append(separator);
            String s = v.getAbsolutePath();
            // DOTO: The following needs more work!
            // Handle paths contains sep
            if (s.indexOf(separator) != -1) {
                if (s.indexOf("\"") != -1) {
                    s = "'" + s + "'";
                }
                else {
                    s = "\"" + s + "\"";
                }
            }
            b.append(s);
        }
        return b.toString();
    }

    private Path setToPath(Hashtable set) {
        Path ret = new Path(getProject());
        for (Enumeration e = set.keys(); e.hasMoreElements();) {
            File v = (File) e.nextElement();
            Path.PathElement el = ret.createPathElement();
            el.setLocation(v);
        }
        return ret;
    }

}


